package MemcachedSOAP;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import org.w3c.dom.NodeList;

import javax.xml.soap.MessageFactory;
import javax.xml.soap.Name;
import javax.xml.soap.Node;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPBodyElement;
import javax.xml.soap.SOAPConnection;
import javax.xml.soap.SOAPConnectionFactory;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFactory;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;

/**
 * A memcached Java SOAP client.
 * 
 * <p>
 * This class is used to access a distant Memcached deamon.<br />
 * The client connects itself to a MemcachedSOAP.cgi webservice hosted on a server somewhere
 * that relays the requests of the client direcly to the Memcached deamon via the Apache::Session::Memorycached interface.
 *
 * <p>
 * This is Open Source software released under the <a href="http://www.gnu.org/licenses/gpl.html">GNU/GPL</a> license.<br />
 * It is provided "as is" without warranty of any kind.<br />
 * Copyright 2007: Nicolas Grevet & Younes Mohmoh, France.<br />
 *
 * <p>
 * Version history:<br />
 * 2007-06-01: first release<br />
 * 2007-06-04: various improvements and fixes.<br />
 * 
 * @author Nicolas GREVET
 * @author Younes MOHMOH
 */

public class MemcachedSOAPClient
{
	private String PREFIX			= "urn";
	private String NAMESPACE_URI	= "MemcachedSOAPClass";
	private String WS_URI			= "http://10.75.201.52/cgi-bin/MemcachedSOAP.cgi";

	/**
	 * Basic constructor without arguments.<br />
	 * <i>Uses the default MemcachedSOAP webservice.</i>
	 * 
	 * @author Younes MOHMOH
	 */

	public MemcachedSOAPClient() 
	{
	}
	
	/**
	 * Basic constructor with arguments.<br />
	 * <i>Can be used to change the default webservice namespace and/or uri.</i>
	 * 
	 * @author Nicolas GREVET
	 * @param prefix The prefix of the namespace (default to <i>urn</i>)
	 * @param namespace The namespace of the webservice (default to <i>MemcachedSOAPClass</i>)
	 * @param service The URI of the webservice
	 */

	public MemcachedSOAPClient(String prefix, String namespace, String service) 
	{
		PREFIX = prefix;
		NAMESPACE_URI = namespace;
		WS_URI = service;
	}

	/**
	 * This method allows a user/tester to specify a proxy host and port.
	 * 
	 * @author Nicolas GREVET
	 * @param host The hostname or IP of the proxy server
	 * @param port The port number of the proxy server
	 */
	
	public void setProxy(String host, String port)
	{
		Properties props = new Properties(System.getProperties());
		props.put("http.proxySet", "true");
		props.put("http.proxyHost", host);
		props.put("http.proxyPort", port);
		Properties newprops = new Properties(props);
		System.setProperties(newprops);
	}

	/**
	 * serverStatusOK <b>checks if the remote SOAP server is up</b> and able to answer to our SOAP requests.
	 * Basicaly, it just sends a SOAP request and see if the server answers to it.
	 * 
	 * @author Nicolas GREVET
	 * @author Younes MOHMOH
	 * @return <b>true</b> if the remote SOAP server is up and able to answer to our SOAP requests or <b>false</b> if it isn't.
	 */
	
	@SuppressWarnings("finally")
	public boolean serverStatusOK()
	{
		SOAPMessage response = null;
		SOAPConnection con = null;
		
		try
		{
			// Cration d'un message SOAP
			SOAPConnectionFactory factory = SOAPConnectionFactory.newInstance();
			con = factory.createConnection();

			MessageFactory mf = MessageFactory.newInstance(); 
			SOAPMessage msg = mf.createMessage();
			//msg.setProperty(SOAPMessage.WRITE_XML_DECLARATION, true);
			SOAPPart sp = msg.getSOAPPart();

			// Cration de l'enveloppe du message SOAP
			SOAPEnvelope env = sp.getEnvelope();
			env.addNamespaceDeclaration(this.PREFIX, this.PREFIX+":"+this.NAMESPACE_URI);

			// Ajout du header au message SOAP
			SOAPHeader hdr = env.getHeader();
			hdr.detachNode();

			// Ajout du Body au message SOAP
			SOAPBody bdy = env.getBody();
			SOAPFactory fac = SOAPFactory.newInstance();
			Name bdyName = fac.createName("status", "ns1", this.NAMESPACE_URI);
			bdy.addBodyElement(bdyName);

			// Envoi du message SOAP au webservice
			URL endPoint = new URL(this.WS_URI);
			response = con.call(msg, endPoint);
		}
		catch (SOAPException e)
		{
			System.out.println("SOAPException :\n"+e+"\n"+e.getMessage());
		}
		catch (MalformedURLException e)
		{
			System.out.println("MalformedURLException :\n"+e+"\n"+e.getMessage());
		}
		finally
		{
			try
			{
				con.close();
				if ((response != null) && (response.getSOAPBody().getElementsByTagName("statusResponse").item(0) instanceof Node))
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			catch (SOAPException e)
			{
				return false;
			}
		}
	}

	/**
	 * getSession sends a query to the remote webservice in order to <b>gather all the informations
	 * associated to a specific session number</b> (passed in parameter).
	 * 
	 * @author Nicolas GREVET
	 * @author Younes MOHMOH
	 * @param session A <b>string</b> containing the number of the session to gather.
	 * @return A <b>HashMap</b> containing pairs of key and values representing the session or an empty HashMap if query failed.
	 * @throws SOAPException 
	 * @throws UnsupportedOperationException 
	 * @throws MalformedURLException 
	 */
	
	public HashMap<String, Object> getSession(String session) throws UnsupportedOperationException, SOAPException, MalformedURLException 
	{
		SOAPMessage response = null;
		Node responseElement;
		HashMap<String, Object> informations = new HashMap<String, Object>();
		SOAPConnection con = null;
		
		try
		{
			SOAPConnectionFactory factory = SOAPConnectionFactory.newInstance();
			con = factory.createConnection();
			
			MessageFactory mf = MessageFactory.newInstance(); 
			SOAPMessage msg = mf.createMessage();
			//msg.setProperty(SOAPMessage.WRITE_XML_DECLARATION, true);
			SOAPPart sp = msg.getSOAPPart();
	
			// Cration de l'enveloppe du Message SOAP
			SOAPEnvelope env = sp.getEnvelope();
			env.addNamespaceDeclaration(PREFIX, PREFIX+":"+NAMESPACE_URI);
	
			// Ajout du Header au message SOAP
			SOAPHeader hdr = env.getHeader();
			hdr.detachNode();
	
			// Ajout du Body au message SOAP
			SOAPBody bdy = env.getBody();
			SOAPFactory fac = SOAPFactory.newInstance();
			Name bdyName = fac.createName("getSession", "ns1", NAMESPACE_URI);
			SOAPBodyElement bdyElement = bdy.addBodyElement(bdyName);
			Name paramName = fac.createName("_session_id");
			SOAPElement requestParam = bdyElement.addChildElement(paramName);
			requestParam.addTextNode(session);
	
			// Envoi du message SOAP au webservice
			URL endpoint = new URL(WS_URI);
			response = con.call(msg, endpoint);
		}
		finally
		{
			// Fermeture de la connexion
			con.close();
		}
		SOAPBody responseBdy = response.getSOAPBody();
		responseElement = (Node) responseBdy.getFirstChild().getFirstChild().getFirstChild();
		
		/*msg.writeTo(System.out);
		System.out.println();
		response.writeTo(System.out);
		System.out.println();
		System.out.println(responseElement);
		*/
		
		do
		{
			if (responseElement == null)
				throw(new SOAPException());

			Object responseValue = responseElement.getValue();
			if (responseElement.getAttributes().item(0).getTextContent().equals("xsd:base64Binary"))
				responseValue = Base64Coder.decodeString((String) responseValue);
			else if (responseElement.getAttributes().item(0).getTextContent().equals("xsd:string"))
				responseValue = (String) responseValue;
			else if (responseElement.getAttributes().item(0).getTextContent().equals("xsd:int"))
				responseValue = Integer.parseInt((String) responseValue);

			informations.put(responseElement.getNodeName(), responseValue);
			//System.out.println(responseElement);
			//System.out.println(responseElement.getNodeName()+" = "+responseElement.getValue());
			responseElement = (Node) responseElement.getNextSibling();
		}
		while (responseElement instanceof Node);

		return informations;
	}

	/**
	 * setSession sends a query to the remote webservice in order to <b>create a new session</b>.
	 * The method is expecting a HashMap parameter containing a succession of key/value pairs
	 * representing the session's informations.
	 * 
	 * @author Nicolas GREVET
	 * @param session A <b>HashMap</b> containing pairs of keys and values representing the session's informations.
	 * @return a <b>string</b> containing the ID of this new session or null if query failed.
	 * @throws SOAPException 
	 * @throws MalformedURLException 
	 */
	
	public String setSession(HashMap session) throws SOAPException, MalformedURLException
	{
		SOAPMessage response = null;
		NodeList responseElements = null;
		Node responseElement = null;
		SOAPConnection con = null;
		
		try
		{
			SOAPConnectionFactory factory = SOAPConnectionFactory.newInstance();
			con = factory.createConnection();
			
			MessageFactory mf = MessageFactory.newInstance(); 
			SOAPMessage msg = mf.createMessage();
			//msg.setProperty(SOAPMessage.WRITE_XML_DECLARATION, true);
			SOAPPart sp = msg.getSOAPPart();
	
			// Cration de l'enveloppe du Message SOAP
			SOAPEnvelope env = sp.getEnvelope();
			env.addNamespaceDeclaration(PREFIX, PREFIX+":"+NAMESPACE_URI);
	
			// Ajout du Header au message SOAP
			SOAPHeader hdr = env.getHeader();
			hdr.detachNode();
	
			// Ajout du Body au message SOAP
			SOAPBody bdy = env.getBody();
			SOAPFactory fac = SOAPFactory.newInstance();
			Name bdyName = fac.createName("setSession", "ns1", NAMESPACE_URI);
			SOAPBodyElement bdyElement = bdy.addBodyElement(bdyName);
			
			// Pour chaque lment de la HashMap, on rcupre la cl et la valeur
			// On construit ensuite une requte SOAP contenant un node pour chaque cl
			// et un node pour chaque valeur aprs chaque node de cl
			
			int i = 3;
			Iterator keys = session.keySet().iterator();
			while (keys.hasNext())
			{
				Object key = keys.next();
				
				Name paramName = fac.createName("c-gensym"+i);
				Name paramAttr = fac.createName("xsi:type");
				SOAPElement requestParam = bdyElement.addChildElement(paramName);
				requestParam.addAttribute(paramAttr, "xsd:string");
				requestParam.addTextNode((String) key);
				i += 2;
				
				Name paramName2 = fac.createName("c-gensym"+i);
				Name paramAttr2 = fac.createName("xsi:type");
				SOAPElement requestParam2 = bdyElement.addChildElement(paramName2);
				
				if (session.get(key) instanceof String)
				{
					requestParam2.addAttribute(paramAttr2, "xsd:base64Binary");
					requestParam2.addTextNode(Base64Coder.encodeString(session.get(key).toString()));
				}
				else if (session.get(key) instanceof Integer)
				{
					requestParam2.addAttribute(paramAttr2, "xsd:int");
					requestParam2.addTextNode(session.get(key).toString());
				}
				
				i += 2;
			}
			
			// Envoi du message SOAP au webservice
			URL endpoint = new URL(WS_URI);
			response = con.call(msg, endpoint);
		}
		finally
		{
			// Fermeture de la connexion
			con.close();
		}
		
		// On regarde ensuite la rponse reue et on renvoie le numro de session s'il existe
		// Sinon on assume que a s'est mal pass, et on renvoie null
		
		SOAPBody responseBdy = response.getSOAPBody();
		responseElements = responseBdy.getElementsByTagName("setSessionResponse");
		
		if ((responseElement = (Node) responseElements.item(0)) instanceof Node)
		{
			return ((Node) responseElement.getFirstChild()).getValue();
			
			// Mthode crade si la mthode normale ne marche plus
			//int index = responseElement.getFirstChild().toString().indexOf(">");
			//String sessionNum = responseElement.getFirstChild().toString().substring(++index);
			//sessionNum = sessionNum.substring(0, 32);
			//return sessionNum;
		}
		else
		{
			return null;
		}
	}
	
	/**
	 * Getter for the namespace of the webservice
	 * @author Nicolas GREVET
	 * @return the webservice's namespace
	 */
	
	public String getNamespace()
	{
		return NAMESPACE_URI;
	}

	/**
	 * Getter for the namespace's prefix of the webservice
	 * @author Nicolas GREVET
	 * @return the prefix of the webservice's namespace
	 */
	
	public String getPrefix()
	{
		return PREFIX;
	}

	/**
	 * Getter for the webservice's URI
	 * @author Nicolas GREVET
	 * @return the webservice URI
	 */
	
	public String getURI()
	{
		return WS_URI;
	}
	
	/**
	 * Setter for the namespace of the webservice
	 * @author Nicolas GREVET
	 * @param namespace the webservice's namespace
	 */
	
	public void setNamespace(String namespace)
	{
		NAMESPACE_URI = namespace;
	}

	/**
	 * Setter for the namespace's prefix of the webservice
	 * @author Nicolas GREVET
	 * @param prefix the prefix of the webservice's namespace
	 */
	
	public void setPrefix(String prefix)
	{
		PREFIX = prefix;
	}

	/**
	 * Getter for the webservice's URI
	 * @author Nicolas GREVET
	 * @param service the webservice URI
	 */
	
	public void setURI(String service)
	{
		WS_URI = service;
	}
}